import * as commonjsHelpers from 'commonjsHelpers';

var pe = 'pe';
var foo = commonjsHelpers.commonjsRequire(`ta${pe}`);
console.log(foo);

var input = {

};

export default input;
export { input as __moduleExports };
