var foo = 1;
var bar = 2;

var foo_1 = 'a';
var bar_1 = 'b';

var input = {
	foo: foo_1,
	bar: bar_1
};

export default input;
export { input as __moduleExports };
export { foo_1 as foo };
export { bar_1 as bar };
